#ifndef __lint
/*@unused@ */
static const char rcsid[] = 
"$Id: align.c,v 1.4 1999/04/16 19:42:58 schwartz Exp $";
#endif

#include "psublast/libc.h"
#include "psublast/types.h"
#include "psublast/misc.h"
#include "psublast/edit.h"
#include "psublast/align.h"

enum { ERROR_FRACTION=2  /* 1/this */
,      MAX_SPACE=1000000
};

#undef ICEIL
#define ICEIL(x,y) ((((x)-1)/(y))+1)

/* ----- pool allocator ----- */
typedef struct space_struct {
  int *space_array;
  int used, size;
} space_t, *space_ptr;

static space_ptr new_space(int MAX_D)
{
  space_ptr p;
  int amount;

  p = ckalloc(sizeof(space_t));
  amount = MAX_SPACE;
  p->space_array = ckalloc(sizeof(int)*amount);
  p->used = 0; p->size = amount;
  return p;
}

static void free_space(space_ptr sp)
{
  sp->space_array = ckfree(sp->space_array);
  sp = ckfree(sp);
}

static int *get_space(space_ptr S, int amount)
{
  int *s = S->space_array+S->used;
  if (S->used+amount > S->size) fatal("out of space");
  S->used += amount;
  return s;
}

/* ----- */

static int get_last(int **flast_d, int d, int diag, int *row1)
{
  if (flast_d[d-1][diag-1] > MAX(flast_d[d-1][diag], flast_d[d-1][diag+1])) {
    *row1 = flast_d[d-1][diag-1];
    return diag-1;
  } 
  if (flast_d[d-1][diag] > flast_d[d-1][diag+1]) {
    *row1 = flast_d[d-1][diag];
    return diag;
  }
  *row1 = flast_d[d-1][diag+1];
  return diag+1;
}

/* Basic O(ND) time, O(N) space, alignment function. 
 * Parameters:
 *   s1, len1        - first sequence and its length
 *   s2, len2        - second sequence and its length
 *   reverse         - direction of alignment
 *   xdrop_threshold -
 *   mismatch_cost   -
 *   e1, e2          - endpoint of the computed alignment
 *   edit_script     -
 */
int align_basic(
	const uchar *s1, int len1,
	const uchar *s2, int len2,
	int reverse, int xdrop_threshold, int match_cost, int mismatch_cost,
	int *e1, int *e2, edit_script_t *S)
{
    int col,			/* column number */
     d,				/* current distance */
     k,				/* current diagonal */
     flower, fupper,            /* boundaries for searching diagonals */
     row,		        /* row number */
     MAX_D, 			/* maximum cost */
     ORIGIN,
     return_val = 0;
    int **flast_d;		/* rows containing the last d */
    int *max_row;		/* reached for cost d=0, ... len1.  */

    int X_pen = xdrop_threshold;
    int M_half = match_cost/2;
    int Op_cost = mismatch_cost + M_half*2;
    int D_diff = ICEIL(X_pen+M_half, Op_cost);

    int *max_row_free, x, cur_max, b_diag = 0, best_diag = INT_MAX/2;
    char nlower = 0, nupper = 0;
    space_ptr space;
    

    MAX_D = (int) (len1/ERROR_FRACTION + 1);
    ORIGIN = MAX_D + 2;
    *e1 = *e2 = 0;

    if (reverse)
	for (row = 0; row < len2 && row < len1 && (s1[len1-1-row] == s2[len2-1-row]); row++)
	    /*empty*/ ;
    else
	for (row = 0; row < len2 && row < len1 && (s1[row] == s2[row]); row++)
	    /*empty*/ ;

    *e1 = row;
    *e2 = row;
    if (row == len1) {
        if (S != NULL)
	  edit_script_rep(S, row);
	/* hit last row; stop search */
	return 0;
    }
    flast_d = ckalloc((MAX_D + 2) * sizeof(int *));
    if (S==NULL) {
      space = 0;
      flast_d[0] = ckalloc((MAX_D + MAX_D + 6) * sizeof(int) * 2);
      flast_d[1] = flast_d[0] + MAX_D + MAX_D + 6;
    } else {
      space = new_space(MAX_D);
      flast_d[0] = get_space(space, 7)-ORIGIN+3;
      flast_d[1] = get_space(space, 9)-ORIGIN+4;
    }
    max_row_free = ckalloc(sizeof(int) * (MAX_D + 1 + D_diff));
    max_row = max_row_free + X_pen/Op_cost;
    for (k = 0; k < D_diff; k++)
	max_row_free[k] = 0;

    flast_d[0][ORIGIN] = row;
    max_row[0] = (row + row)*M_half;

    flower = ORIGIN - 1;
    fupper = ORIGIN + 1;

    d = 1;
    while (d <= MAX_D) {
	int fl0, fu0;
	flast_d[d - 1][flower - 1] = flast_d[d - 1][flower] = -1;
	flast_d[d - 1][fupper] = flast_d[d - 1][fupper + 1] = -1;
	x = max_row[d - D_diff] + Op_cost * d - X_pen;
	x = ICEIL(x, M_half);	
	cur_max = 0;
	fl0 = flower;
	fu0 = fupper;
	for (k = fl0; k <= fu0; k++) {
	    /* get space for the next edit instruction */
	    row = MAX(flast_d[d - 1][k + 1], flast_d[d - 1][k]) + 1;
	    row = MAX(row, flast_d[d - 1][k - 1]);
	    col = row + k - ORIGIN;
	    if (row + col >= x)
		fupper = k;
	    else {
		if (k == flower)
		    flower++;
		else
		    flast_d[d][k] = -1;
		continue;
	    }
	    /* slide down the diagonal */
	    if (reverse) {
		while (row < len1 && col < len2 && s1[len1-1-row] == s2[len2-1-col]) {
		    ++row;
		    ++col;
		}
	    } else {
		while (row < len1 && col < len2 && s1[row] == s2[col]) {
		    ++row;
		    ++col;
		}
	    }
	    flast_d[d][k] = row;
	    if (row + col > cur_max) {
		cur_max = row + col;
		b_diag = k;
	    }
	    if (row == len1) {
		flower = k+1; nlower = 1;
	    }
	    if (col == len2) {
		fupper = k-1; nupper = 1;
	    }
	}
	k = cur_max*M_half - d * Op_cost;
	if (max_row[d - 1] < k) {
	    max_row[d] = k;
	    return_val = d;
	    best_diag = b_diag;
	    *e1 = flast_d[d][b_diag];
	    *e2 = (*e1)+b_diag-ORIGIN;
	} else {
	    max_row[d] = max_row[d - 1];
	}
	if (flower > fupper)
	    break;
	d++;
	if (!nlower) flower--; 
	if (!nupper) fupper++;
	if (S==NULL)
	  flast_d[d] = flast_d[d - 2];
	else 
	  flast_d[d] = get_space(space, fupper-flower+5)-flower+2;
    }

    /*printf("return\n"); */
    if (S!=NULL) { /*trace back*/
      int row1, col1, diag1, diag;
      d = return_val; diag = best_diag;
      row = *e1; col = *e2;
      while (d > 0) {
	diag1 = get_last(flast_d, d, diag, &row1);
	col1 = row1+diag1-ORIGIN;
	if (diag1 == diag) {
	  if (row-row1 > 0) edit_script_rep(S, row-row1);
	} else if (diag1 < diag) {
	  if (row-row1 > 0) edit_script_rep(S, row-row1);
	  edit_script_ins(S,1);
	} else {
	  if (row-row1-1> 0) edit_script_rep(S, row-row1-1);
	  edit_script_del(S, 1);
	}
	d--; diag = diag1; col = col1; row = row1;
      }
      edit_script_rep(S, flast_d[0][ORIGIN]);
      if (!reverse) 
	edit_script_reverse_inplace(S);
      free_space(space);
    } else {
      free(flast_d[0]);
    }
    free(max_row_free);
    free(flast_d);
    return return_val;
}
