#ifndef __lint
/*@unused@ */
static const char rcsid[] = 
"$Id: galign.c,v 1.6 1999/06/23 18:42:20 zzhang Exp $";
#endif

#include "psublast/libc.h"
#include "psublast/types.h"
#include "psublast/misc.h"
#include "psublast/edit.h"
#include "psublast/align.h"

enum {         /* half of the (fixed) match score */
       ERROR_FRACTION=2  /* 1/this */
,      MAX_SPACE=1000000
,      sC = 0, sI = 1, sD = 2, LARGE=100000000
};

#undef ICEIL
#define ICEIL(x,y) ((((x)-1)/(y))+1)

/* ----- pool allocator ----- */
typedef struct three {
  int I, C, D;
} three_val;

typedef struct space_struct {
  three_val *space_array;
  int used, size;
} space_t, *space_ptr;

static space_ptr new_space(int MAX_D)
{
  space_ptr p;
  int amount;

  p = ckalloc(sizeof(space_t));
  amount = MAX_SPACE;
  p->space_array = ckalloc(sizeof(three_val)*amount);
  p->used = 0; p->size = amount;
  return p;
}

static void free_space(space_ptr sp)
{
  sp->space_array = ckfree(sp->space_array);
  sp = ckfree(sp);
}

static three_val *get_space(space_ptr S, int amount)
{
  three_val *s = S->space_array+S->used;
  if (amount < 0) return NULL;  
  if (S->used+amount > S->size) fatal("out of space");
  S->used += amount;
  return s;
}

/* ----- */

static int gcd(int a, int b)
{
  int c;
  if (a < b) {
    c = a;
    a = b; b = c;
  }
  while ((c = a % b) != 0) {
    a = b; b = c;
  }
  return b;
}


static int gdb3(int *a, int *b, int *c)
{
  int g;
  if (*b == 0) g = gcd(*a, *c);
  else g = gcd(*a, gcd(*b, *c));
  if (g > 1) {
    *a /= g;
    *b /= g;
    *c /= g;
  }
  return g;
}

static int get_lastC(three_val **flast_d, int *lower, int *upper, int *d, int diag, int Mis_cost, int *row1)
{
  int row;
  
  if (diag >= lower[(*d)-Mis_cost] && diag <= upper[(*d)-Mis_cost]) {
    row = flast_d[(*d)-Mis_cost][diag].C;
    if (row >= MAX(flast_d[*d][diag].I, flast_d[*d][diag].D)) {
      *d = *d-Mis_cost;
      *row1 = row;
      return sC;
    }
  }
  if (flast_d[*d][diag].I > flast_d[*d][diag].D) {
    *row1 = flast_d[*d][diag].I;
    return sI;
  } else {
    *row1 = flast_d[*d][diag].D;
    return sD;
  }
}

static int get_last_ID(three_val **flast_d, int *lower, int *upper, int *d, int diag, int GO_cost, int GE_cost, int IorD)
{
  int ndiag; 
  int row;
  if (IorD == sI) { ndiag = diag -1;} else ndiag = diag+1;
  if (ndiag >= lower[(*d)-GE_cost] && ndiag <=upper[(*d)-GE_cost]) 
    row = (IorD == sI)? flast_d[(*d)-GE_cost][ndiag].I: flast_d[(*d)-GE_cost][ndiag].D;
  else row = -100;
  if (ndiag >= lower[(*d)-GO_cost-GE_cost] && ndiag <= upper[(*d)-GO_cost-GE_cost] && row < flast_d[(*d)-GO_cost-GE_cost][ndiag].C) {
    *d = (*d)-GO_cost-GE_cost;
    return sC;
  }
  *d = (*d)-GE_cost;
  return IorD;
}

static int get_lastI(three_val **flast_d, int *lower, int *upper, int *d, int diag, int GO_cost, int GE_cost)
{
  return get_last_ID(flast_d, lower, upper, d, diag, GO_cost, GE_cost, sI);
}

		     
static int get_lastD(three_val **flast_d, int *lower, int *upper, int *d, int diag, int GO_cost, int GE_cost)
{
  return get_last_ID(flast_d, lower, upper, d, diag, GO_cost, GE_cost, sD);
}

/* Basic O(ND) time, O(N) space, alignment function. 
 * Parameters:
 *   s1, len1        - first sequence and its length
 *   s2, len2        - second sequence and its length
*   reverse         - direction of alignment
 *   xdrop_threshold -
 *   mismatch_cost   -
 *   e1, e2          - endpoint of the computed alignment
 *   edit_script     -
 */

int gap_align_basic(
	const uchar *s1, int len1,
	const uchar *s2, int len2,
	int reverse, int xdrop_threshold, 
	int match_score,
	int mismatch_score,
	int gap_open,
	int gap_extend,
	int *e1, int *e2, edit_script_t *S)
{
    int col,			/* column number */
     d,				/* current distance */
     k,				/* current diagonal */
     flower, fupper,            /* boundaries for searching diagonals */
     row,		        /* row number */
     MAX_D, 			/* maximum cost */
     ORIGIN,
     return_val = 0;
    three_val **flast_d;		/* rows containing the last d */
    int *max_row;		/* reached for cost d=0, ... len1.  */
    int Mis_cost, GO_cost, GE_cost;
    int D_diff, gd;
    int M_half;
    int max_cost;
    int *lower, *upper;

    int *max_row_free, x, cur_max, b_diag = 0, best_diag = INT_MAX/2;
    char nlower = 0, nupper = 0;
    space_ptr space;
    int stop_condition;
    int max_d;
    int *uplow_free;

    if (match_score % 2 == 1) {
      match_score *= 2;
      mismatch_score *= 2;
      xdrop_threshold *= 2;
      gap_open *= 2;
      gap_extend *= 2;
    }
      
    M_half = match_score/2;
    if (gap_open == 0 && gap_extend == mismatch_score+M_half) {
      return align_basic(s1, len1, s2, len2, reverse, xdrop_threshold,
			 match_score, mismatch_score, e1, e2, S);
    }
    Mis_cost = mismatch_score + match_score;
    GO_cost = gap_open;
    GE_cost = gap_extend+M_half;
    gd = gdb3(&Mis_cost, &GO_cost, &GE_cost);
    D_diff = ICEIL(xdrop_threshold+M_half, gd);


    MAX_D = (int) (len1/ERROR_FRACTION + 1);
    max_d = MAX_D*GE_cost;
    ORIGIN = MAX_D + 2;
    max_cost = MAX(Mis_cost, GO_cost+GE_cost);
    *e1 = *e2 = 0;

    if (reverse)
	for (row = 0; row < len2 && row < len1 && (s1[len1-1-row] == s2[len2-1-row]); row++)
	    /*empty*/ ;
    else
	for (row = 0; row < len2 && row < len1 && (s1[row] == s2[row]); row++)
	    /*empty*/ ;

    *e1 = row;
    *e2 = row;
    if (row == len1 || row == len2) {
        if (S != NULL)
	  edit_script_rep(S, row);
	/* hit last row; stop search */
	return row*match_score;
    }
    flast_d = ckalloc((MAX(max_d, max_cost) + 2) * sizeof(three_val *));
    if (S==NULL) {
      space = 0;
      flast_d[0] = ckalloc((MAX_D + MAX_D + 6) * sizeof(three_val) * (max_cost+1));
      for (k = 1; k <= max_cost; k++)
	flast_d[k] = flast_d[k-1] + MAX_D + MAX_D + 6;
    } else {
      space = new_space(MAX_D);
      flast_d[0] = get_space(space, 1)-ORIGIN;
      flast_d[1] = get_space(space, 3)-ORIGIN+1;
    }
    max_row_free = ckalloc(sizeof(int) * (max_d + 1 + D_diff));
    max_row = max_row_free + D_diff;
    for (k = 0; k < D_diff; k++)
	max_row_free[k] = 0;
    uplow_free = ckalloc(sizeof(int)*2*(max_d+1+max_cost));
    lower = uplow_free;
    upper = uplow_free+max_d+1+max_cost;
    /* next 3 lines set boundary for -1,-2,...,-max_cost+1*/
    for (k = 0; k < max_cost; k++) {lower[k] =LARGE;  upper[k] = -LARGE;}
    lower += max_cost;
    upper += max_cost; 

    flast_d[0][ORIGIN].C = row;
    flast_d[0][ORIGIN].I = flast_d[0][ORIGIN].D = -2;
    max_row[0] = (row + row)*M_half;
    lower[0] = upper[0] = ORIGIN;

    flower = ORIGIN - 1;
    fupper = ORIGIN + 1;

    d = 1;
    stop_condition = 1;
    while (d <= max_d) {
	int fl0, fu0;
	x = max_row[d - D_diff] + gd * d - xdrop_threshold;
	x = ICEIL(x, M_half);
	if (x < 0) x=0;
	cur_max = 0;
	fl0 = flower;
	fu0 = fupper;
	for (k = fl0; k <= fu0; k++) {
	    row = -2;
	    if (k+1 <= upper[d-GO_cost-GE_cost] && k+1 >= lower[d-GO_cost-GE_cost]) 
	      row = flast_d[d-GO_cost-GE_cost][k+1].C;
	    if (k+1  <= upper[d-GE_cost] && k+1 >= lower[d-GE_cost] &&
		row < flast_d[d-GE_cost][k+1].D) 
	      row = flast_d[d-GE_cost][k+1].D;
	    row++;
	    if (row+row+k-ORIGIN >= x) 
	      flast_d[d][k].D = row;
	    else flast_d[d][k].D = -2;
	    row = -1; 
	    if (k-1 <= upper[d-GO_cost-GE_cost] && k-1 >= lower[d-GO_cost-GE_cost]) 
	      row = flast_d[d-GO_cost-GE_cost][k-1].C;
	    if (k-1  <= upper[d-GE_cost] && k-1 >= lower[d-GE_cost] &&
		row < flast_d[d-GE_cost][k-1].I) 
	      row = flast_d[d-GE_cost][k-1].I;
	    if (row+row+k-ORIGIN >= x) 
	      flast_d[d][k].I = row;
	    else flast_d[d][k].I = -2;
		
	    row = MAX(flast_d[d][k].I, flast_d[d][k].D);
	    if (k <= upper[d-Mis_cost] && k >= lower[d-Mis_cost]) 
	      row = MAX(flast_d[d-Mis_cost][k].C+1,row);

	    col = row + k - ORIGIN;
	    if (row + col >= x)
		fupper = k;
	    else {
		if (k == flower)
		    flower++;
		else
		    flast_d[d][k].C = -2;
		continue;
	    }
	    /* slide down the diagonal */
	    if (reverse) {
		while (row < len1 && col < len2 && s1[len1-1-row] == s2[len2-1-col]) {
		    ++row;
		    ++col;
		}
	    } else {
		while (row < len1 && col < len2 && s1[row] == s2[col]) {
		    ++row;
		    ++col;
		}
	    }
	    flast_d[d][k].C = row;
	    if (row + col > cur_max) {
		cur_max = row + col;
		b_diag = k;
	    }
	    if (row == len1) {
		flower = k; nlower = k+1;
	    }
	    if (col == len2) {
		fupper = k; nupper = k-1;
	    }
	}
	k = cur_max*M_half - d * gd;
	if (max_row[d - 1] < k) {
	    max_row[d] = k;
	    return_val = d;
	    best_diag = b_diag;
	    *e1 = flast_d[d][b_diag].C;
	    *e2 = (*e1)+b_diag-ORIGIN;
	} else {
	    max_row[d] = max_row[d - 1];
	}
	if (flower <= fupper) {
	  stop_condition++;
	  lower[d] = flower; upper[d] = fupper;
	} else {
	  lower[d] = LARGE; upper[d] = -LARGE;
	}
	if (lower[d-max_cost] <= upper[d-max_cost]) stop_condition--;
	if (stop_condition == 0) break;
	d++;
	flower = MIN(lower[d-Mis_cost], MIN(lower[d-GO_cost-GE_cost], lower[d-GE_cost])-1);
	if (nlower) flower = MAX(flower, nlower);
	fupper = MAX(upper[d-Mis_cost], MAX(upper[d-GO_cost-GE_cost], upper[d-GE_cost])+1);
	if (nupper) fupper = MIN(fupper, nupper);
	if (S==NULL) {
	  if (d > max_cost)
	    flast_d[d] = flast_d[d - max_cost-1];
	} else 
	  flast_d[d] = get_space(space, fupper-flower+1)-flower;
    }

    /*printf("return\n"); */
    if (S!=NULL) { /*trace back*/
      int row1, diag, state;
      d = return_val; diag = best_diag;
      row = *e1; state = sC;
      while (d > 0) {
	if (state == sC) {
	  /* diag will not be changed*/
	  state = get_lastC(flast_d, lower, upper, &d, diag, Mis_cost, &row1);
	  if (row-row1 > 0) edit_script_rep(S, row-row1);
	  row = row1;
	} else {
	  if (state == sI) {
	    /*row unchanged */
	    state = get_lastI(flast_d, lower, upper, &d, diag, GO_cost, GE_cost);
	    diag--;
	    edit_script_ins(S,1);
	  } else {
	    edit_script_del(S,1);
	    state = get_lastD(flast_d, lower, upper, &d, diag, GO_cost, GE_cost);
	    diag++;
	    row--;
	  }
	}
      }
      edit_script_rep(S, flast_d[0][ORIGIN].C);
      if (!reverse) 
	edit_script_reverse_inplace(S);
      free_space(space);
    } else {
      free(flast_d[0]);
    }
    return_val = max_row[return_val];
    free(max_row_free);
    free(flast_d);
    free(uplow_free);
    return return_val;
}
