#ifndef __lint
static const char rcsid[] =
"$Id: mb.c,v 1.46 1999/09/24 20:38:27 schwartz Exp $";
#endif

#include "psublast/libc.h"
#include "psublast/psublast.h"
#include "db.h"
#include "mb.h"
#if 0
#include "mbz.h"
#endif

enum {
    ESTACK_SIZE = 5000
};

struct stack {
    int diag, level, length;
};

struct mb_table {
    unsigned char *seq;
    int slen;

    int W;		/* width */
    int K;		/* blastz threshold */
    int X;		/* xdrop condition */
    int k;		/* gap_align_cutoff */
    int x;		/* xdrop_gapfree */
    int lpm;		/* length of perfect match */
    int surfeit;	/* of 12mer entries */
    ss_t ss;
    argv_scores_t ds;

    int HASH_SIZE;
    int *hashtab;
    long *next_pos;
    ulong mask;

    int stack_index;
    struct stack estack[ESTACK_SIZE];

    int numMSPs;
    gal_t *gal_list;
};

static void sieve_mbt(mbt_t *);
static int extend_hit(mbt_t *, int, int, int);
static int binary_search(int, long*, int);
static gal_t *search(mbt_t *, uchar*, int, int, long*);
static void add_gal(mbt_t *, int, int, int);
static void bld_table(mbt_t *);
static int gap_free_extend(mbt_t *, gal_t*, const uchar *, int, const uchar *, int);

/* normalize -- transform reverse compliment coord to normal coord */
static int normalize(int want, int state, int len, int x)
{
	return (want && state) ? len-x : x; /* len+1? */
}

void free_gal_list(gal_t *list)
{
	while (list) {
		gal_t *p = list->next;	
		ckfree(list);
		list = p;
	}
}

#if 0
static void mb_bz(mbt_t *mbt,
		  uchar *s, int len,
		  int index,
		  qdb_t *qdb,
		  gal_t *mq,
		  int N,
		  char *def, int nid)
{
    gal_t *mp;

    for (mp = mq; mp; mp = mp->next) {
	int i;
	mp->prune = 0;
	i = mp->id = binary_search(mp->b2, qdb->beg, index);
	mp->b2 -= qdb->beg[i];
	mp->e2 -= qdb->beg[i];
	if (gap_free_extend(mbt, mp,
			    s, len,
			    qdb_seq(qdb,i), qdb_length(qdb,i)) == 0)
	    mp->prune = 1;

    }
    for (mp = mq; mp; mp = mp->next) {
	if (mp->prune == 0) {
	    int i = mp->id;
	    print_bz_result(nid, def, qdb->hdr[i]);
	    call_blastz_extend_msps(
		mp, N,
                s, len,
                qdb_seq(qdb,i), qdb_length(qdb,i),
                &mbt->ds, mbt->ss, mbt->X, mbt->lpm*4, mbt->K
            );
	}
    }
}
#endif

void mb_search(mbt_t *mbt,
	       uchar* t, int len,
	       int index, qdb_t *qdb,
	       char *defline,
	       align_params_t *P)
{
    long N;
    gal_t *gal_orig;
    char def[HDRSAVE];
    const int nid = P->idlen;
    const int prog_flag = P->prog_flag;
    const int xdrop = P->xdrop_ext;
    const int mismatch_penalty = - P->mismatch_penalty_mb;
    const int match_score = P->match_score;
    const int gop = P->gap_open_penalty;
    const int gep = P->gap_extension_penalty;

    fmt_id(def, '>', defline);
    gal_orig = search(mbt, t, len, prog_flag, &N);
    if (N > 0) {
        uchar *s = readdb_unpack(t, len);
	if (prog_flag == MB_DO_BZ) {
	    fatal("blatz is no longer supported!");
#if 0
	    mb_bz(mbt, s, len, index, qdb, gal_orig, N, def, nid);
	    s = ckfree(s);
            free_gal_list(gal_orig);
#endif
	} else {
	    gal_t *gp;
            gal_t *gal_new = 0;
            for (gp = gal_orig; gp;) {
                gal_t *ngp;
                int i = binary_search(gp->b2, qdb->beg, index);
                if (gp->b2 < qdb->beg[i])
        	    fprintf(stderr,
        	      "mb: warning %d %ld %d\n", gp->b2, qdb->beg[i], i);
                gp->id = i;
                gp->b2 -= qdb->beg[i];
                gp->e2 -= qdb->beg[i];
                /*debugf("%d %d %d %d\n", gp->b1, gp->b2, gp->e1, gp->e2); */
                for (ngp = gal_new; ngp; ngp = ngp->next) {
        	    if (ngp->id == i)
        	        if (GAL_CONTAINS(ngp, gp) && GAL_CLOSE(ngp, gp))
        	            break;
                }
                if (ngp != 0) {
		    /* unlink and free */
		    gal_t *t = gp->next;
        	    ckfree(gp);
		    gp = t;
                } else {
        	    int j = (P->output_al != MB_ALIGN_NONE ?
				 mb_extend_gap_align_script :
				 mb_extend_gap_align) \
			(gp, xdrop, match_score,
			 mismatch_penalty, gop, gep,
			 s, len,
			 qdb_seq(qdb,i), qdb_length(qdb,i));

        	    gp->prune = 0;
        	    gp->diag = j;
        	    for (ngp = gal_new; ngp; ngp = ngp->next) {
        	        if (ngp->prune == 0)
        		    if (ngp->id == i)
        		        if (GAL_CONTAINS(gp, ngp) && GAL_CLOSE(ngp, gp))
        			    ngp->prune = 1;
        	    }
        	    ngp = gp->next;
        	    gp->next= gal_new;
        	    gal_new = gp;
        	    gp = ngp;
                }
            }
            /*prnt_msps(gal_new, qdb, P, def); */
            for (gp = gal_new; gp; gp = gp->next) {
                if (gp->prune == 0) {
        	    int len = qdb_length(qdb,gp->id);
        	    int cmpl = qdb_is_revcomp(qdb, gp->id);
        	    int doit = P->normalize_revcomp_flag;
        	    int b1 = gp->b1;
        	    int e1 = gp->e1;
        	    int b2 = normalize(doit, cmpl, len, gp->b2);
        	    int e2 = normalize(doit, cmpl, len, gp->e2);

		    if (P->output_al == MB_ALIGN_LAV) putchar('#');
        	    print_mb_result(nid, def, qdb->hdr[gp->id],
                            b1, b2, e1, e2, gp->diag, ' ');
		    if (gp->S) {
			if (P->output_al == MB_ALIGN_LAV) {
			    putchar('\n');
		            print_align_lav(gp->diag, s, qdb_seq(qdb,gp->id), 
				      gp->b1, gp->e1, gp->b2, gp->e2, gp->S);
			} else {
			    putchar(' ');
			    print_align_summary(gp->diag, s, qdb_seq(qdb,gp->id),
                                      gp->b1, gp->e1, gp->b2, gp->e2, gp->S);
			}
			gp->S = edit_script_free(gp->S);
		    } else {
			putchar('\n');
		    }
                }
            }
            s = ckfree(s);
            free_gal_list(gal_new);
	}
    }
    return;
}

int est_search(mbt_t *mbt, unsigned char *t, int l2, align_params_t *P)
{
    long i;
    return search(mbt, t, l2, P->prog_flag, &i) != 0;
}

static void bld_table(mbt_t *mbt)
{
    const signed char *const encoding = fasta_encoding;
    register ulong ecode;
    register ulong mask;
    int j, h;
    register int i;
    register unsigned char *t, *p;
    uchar *s;

    /* perform initializations */
    mbt->HASH_SIZE = (1 << (mbt->W*8)) - 1;
    mask = mbt->mask = (1 << (mbt->W*8 - 2)) - 1;
    if (mbt->hashtab == 0) {
	mbt->hashtab = ckalloc(sizeof(int *) * (mbt->HASH_SIZE + 1));
	mbt->next_pos = ckalloc(sizeof(long *) * BLOCKSIZE);
    }
    for (i = 0; i < mbt->HASH_SIZE + 1; i++)
	mbt->hashtab[i] = 0;

    s = mbt->seq;
    t = s - 1;
    ecode = 0L;
    for (i = 1; i < mbt->W*4; ++i)
	ecode = (ecode << 2) + (unsigned long)encoding[*++t];
    p = s;
    while (*++t) {
	ulong e = encoding[*t];
	if (e == ~0U) {
	    ecode = 0;
	    p = t + mbt->W*4;
	    *t = toupper(*t);
	} else {
	    ecode = ((ecode & mask) << 2) + e;
	    if (t >= p) {
		mbt->next_pos[i] = mbt->hashtab[ecode];
		mbt->hashtab[ecode] = i;
	    }
	}
	i++;
    }
    mbt->mask = (1 << (mbt->W*8 - 8)) - 1;
    sieve_mbt(mbt);
}

static void sieve_mbt(mbt_t *mbt)
{
    const unsigned char *const code = fasta_decoding;
    int i;
    for (i = 0; i <= mbt->HASH_SIZE; i++) {
	int h;
	int j = 0;
	for (h = mbt->hashtab[i]; h; h = mbt->next_pos[h])
	    j++;
	if (j > mbt->surfeit) {
	    fprintf(stderr, "%d ", j);
	    for (j = 0; j < mbt->W*4; j++) {
		fprintf(stderr, "%c", code[(i >> (mbt->W*8 - 2 - j*2)) % 4]);
	    }
	    fprintf(stderr, "\n");
	    mbt->hashtab[i] = 0;
	}
    }

}

static int extend_hit(mbt_t *mbt, int i, int p, int prog_flag)
{
    int j, len = mbt->W*4;
    const int EST = prog_flag == MB_DO_EST;

    for (j = 0; j < mbt->stack_index;) {
	if (mbt->estack[j].diag == i - p) {
	    if (mbt->estack[j].level < i - 4) {
		if (!EST)
		    add_gal(mbt, mbt->estack[j].level, mbt->estack[j].diag, mbt->estack[j].length);
		mbt->estack[j].length = len;
		mbt->estack[j].level = i;
		return 0;
	    }
	    mbt->estack[j].length += 4;
	    if (EST)
		if (mbt->estack[j].length > mbt->lpm*4)
		    return 1;
	    mbt->estack[j].level = i;
	    return 0;
	} else if (mbt->estack[j].level >= i - 4)
	    j++;
	else {
	    mbt->stack_index--;
	    if (!EST)
		add_gal(mbt, mbt->estack[j].level, mbt->estack[j].diag, mbt->estack[j].length);
	    mbt->estack[j] = mbt->estack[mbt->stack_index];
	}
    }
    if (mbt->stack_index >= ESTACK_SIZE)
	fatal("stack overflow in extend_hit");

    mbt->estack[mbt->stack_index].diag = i - p;
    mbt->estack[mbt->stack_index].level = i;
    mbt->estack[mbt->stack_index].length = len;
    mbt->stack_index++;
    return 0;
}

static int binary_search(int n, long A[], int i)
{
    int m, b, e;

    b = 0;
    e = i;
    while (b < e - 1) {
	m = (b + e) / 2;
	if (A[m] > n)
	    e = m;
	else
	    b = m;
    }
    return b;
}

static void add_gal(mbt_t *mbt, int level, int diag, int length)
{
    gal_t *mp;

    if (length <= mbt->lpm*4)
	return;

    mp = ckalloc(sizeof(*mp)); 
    mp->b1 = level - length + 1;
    mp->e1 = level;
    mp->b2 = mp->b1 - diag;
    mp->e2 = mp->e1 - diag;
    mp->length = length;

    mp->next = mbt->gal_list;
    mbt->gal_list = mp;

    mbt->numMSPs++;
}

static gal_t *search(mbt_t *mbt, uchar* s, int len, int prog_flag, long N[])
{
    register int h;
    register uchar* t;
    register ulong ecode;
    register ulong mask;
    int i;
    const int EST = prog_flag == MB_DO_EST;

    mbt->numMSPs = 0;
    mbt->gal_list = 0;
    t = s - 1;
    mbt->stack_index = 0;
    ecode = 0L;
    for (i = 0; i < (mbt->W - 1)*4; i += 4) {
	ecode = (ecode << 8) + *++t;
    }
    i += 4;
    mask = mbt->mask;
    while (i < len) {
	ecode = ((ecode & mask) << 8) + *++t;
	for (h = mbt->hashtab[ecode]; h; h = mbt->next_pos[h])
	    if (!EST)
		(void) extend_hit(mbt, i, h, prog_flag);
	    else if (extend_hit(mbt, i, h, prog_flag))
		return (gal_t *) 1; /* XXX */
	i += 4;
    }

    if (EST) {
	return 0;
    } else {
	for (i = 0; i < mbt->stack_index; i++)
	    add_gal(mbt, mbt->estack[i].level, mbt->estack[i].diag, mbt->estack[i].length);

	*N = mbt->numMSPs;
	return mbt->gal_list;
    }
}

mbt_t *mb_init(void)
{
    mbt_t *mbt = ckalloc(sizeof(*mbt));
    mbt->seq = 0;
    mbt->slen = 0;

    mbt->W = 3;
    mbt->lpm = 20;
    mbt->surfeit = 0;

    mbt->numMSPs = 0;
    mbt->gal_list = 0;

    mbt->HASH_SIZE = (1 << 24) - 1;
    mbt->hashtab = 0;
    mbt->next_pos = 0;
    mbt->mask = 0;
    mbt->stack_index = 0;

    return mbt;
}

void mb_start(mbt_t *mbt, unsigned char *dna, int len, int w, int p)
{
    mbt->seq = dna;
    mbt->slen = len;
    mbt->lpm = w;
    mbt->surfeit = p;
    bld_table(mbt);
}

void mb_end(mbt_t *mbt)
{
    mbt->hashtab = ckfree(mbt->hashtab);
    mbt->next_pos = ckfree(mbt->next_pos);
    mbt = ckfree(mbt);
}

void mb_blastz_start(mbt_t *mbt, int X, int K, int k, int x)
{
    DNA_scores(&(mbt->ds), mbt->ss);
    mbt->X = X;
    mbt->K = K;
    mbt->k = k;
    mbt->x = x;
}


static int gap_free_extend(mbt_t *mbt, gal_t *mp, const uchar *seq1, int len1, const uchar *seq2, int len2)
{
    const int X = mbt->x;
    const uchar *s, *q;
    int left_sum, right_sum, sum, score;
    const uchar *const seq1_end = seq1 + len1;
    const uchar *const seq2_end = seq2 + len2;

    score = (mp->e1 - mp->b1 + 1)*mbt->ss['A']['A'];
    if (score > mbt->k)
	return 1;
    score = mbt->k - score;
    right_sum = sum = 0;
    q = seq1 + mp->e1;
    s = seq2 + mp->e2;
    while (s < seq2_end && q < seq1_end && sum >= right_sum - X)
	if ((sum += mbt->ss[*s++][*q++]) > right_sum) {
	    right_sum = sum;
	    if (right_sum > score)
		return 1;
	}
    score -= right_sum;
    /* extend to the left */
    left_sum = sum = 0;
    q = seq1 + mp->b1;
    s = seq2 + mp->b2;
    while (s > seq2 && q > seq1 && sum >= left_sum - X)
	if ((sum += mbt->ss[*--s][*--q]) > left_sum) {
	    left_sum = sum;
	    if (left_sum > score)
		return 1;
	}
    return 0;
}
